package at.gv.egiz.moazs.backend;

import at.gv.egiz.moazs.repository.DeliveryRepository;
import at.gv.egiz.moazs.scheme.Marshaller;
import at.gv.egiz.moazs.scheme.MsgResponse;
import org.apache.commons.io.FileUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.concurrent.CompletableFuture;

import static at.gv.egiz.moazs.MoaZSException.moaZSException;
import static java.lang.String.format;
import static java.util.concurrent.CompletableFuture.allOf;
import static java.util.concurrent.CompletableFuture.supplyAsync;

@Component
public class SaveResponseToFileSink {

    private static final Logger log = LoggerFactory.getLogger(SaveResponseToFileSink.class);
    private static final String SAVING_FAILED_MSG = "Could not save response with AppDeliveryId=%s.";
    private static final String SAVE_RESPONSE_MSG = "Saving response to {}..." ;

    private final SimpleDateFormat isoFormatter;
    private final Marshaller msgMarshaller;
    private final DeliveryRepository repository;

    @Autowired
    public SaveResponseToFileSink(Marshaller msgMarshaller, DeliveryRepository repository) {
        this.msgMarshaller = msgMarshaller;
        this.repository = repository;
        this.isoFormatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
    }

    /**
     * Save response and it's binary version to the file system.
     *
     * @param response
     * @return Future that completes when both responses have been written to the file system.
     */
    public CompletableFuture<Void> save(MsgResponse response, String rootPath) {

        var responseID = response.getResponseID();
        log.info("Save msg:{} with AppDeliveryID={}.",
                response.getRootElementLocalPart(), response.getAppDeliveryID());

        var responsePath = generatePath(rootPath, responseID, "xml");
        var storeResponseToFileSystemFuture = supplyAsync(() -> msgMarshaller.marshallXml(response.getResponseAsJAXBElement()))
                .thenApply(responseString -> responseString.getBytes(StandardCharsets.UTF_8))
                .thenAccept(responseByteArray -> storeToFile(responsePath, responseByteArray))
                .exceptionally(ex -> logException(ex, responseID));

        var binaryResponsePath = generatePath(rootPath, responseID, "binary.xml");
        var storeBinaryResponseToFileSystemFuture = supplyAsync(() -> repository.retrieveBinaryResponse(responseID).get())
                .thenAccept(binaryResponseByteArray -> storeToFile(binaryResponsePath, binaryResponseByteArray))
                .exceptionally(ex -> logException(ex, responseID));

        return allOf(storeResponseToFileSystemFuture, storeBinaryResponseToFileSystemFuture);

    }

    private String generatePath(String rootPath, String id, String suffix) {
        var folder = sanitizeFileString(id);
        var nowInIso8601 = isoFormatter.format(new Date());
        return format("%s/%s/%s.%s", rootPath, folder, nowInIso8601,  suffix);
    }

    private String sanitizeFileString(String fileString) {
        return fileString.replaceAll("[^a-zA-Z0-9\\._\\-]", "");
    }

    private Void logException(Throwable ex, String appDeliveryID) {
        if(log.isErrorEnabled()) {
            log.error(format(SAVING_FAILED_MSG, appDeliveryID), ex);
        }
        return null;
    }

    private void storeToFile(String path, byte[] content) {
        log.trace(SAVE_RESPONSE_MSG, path);
        try {
            FileUtils.writeByteArrayToFile(new File(path), content);
        } catch (IOException e) {
            throw moaZSException(e.getMessage(), e);
        }
    }

}
